@@ -1,4 +1,4 @@
-# Copyright 2004-2015 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2017 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,9 +22,11 @@
 # This file contains code for initializing and managing the display
 # window.
 
+from __future__ import print_function
 import renpy.display
 import renpy.audio
 import renpy.text
+import renpy.test
 
 import pygame_sdl2 as pygame
 
@@ -33,21 +35,22 @@ import os
 import time
 import cStringIO
 import threading
+import copy
 
 import_time = time.time()
 
 try:
-    import android # @UnresolvedImport
+    import android  # @UnresolvedImport
 except:
     android = None
 
 TIMEEVENT = pygame.event.register("TIMEEVENT")
-PERIODIC =  pygame.event.register("PERIODIC")
+PERIODIC = pygame.event.register("PERIODIC")
 REDRAW = pygame.event.register("REDRAW")
 EVENTNAME = pygame.event.register("EVENTNAME")
 
 # All events except for TIMEEVENT and REDRAW
-ALL_EVENTS = set(pygame.event.get_standard_events()) # @UndefinedVariable
+ALL_EVENTS = set(pygame.event.get_standard_events())  # @UndefinedVariable
 ALL_EVENTS.add(PERIODIC)
 ALL_EVENTS.add(EVENTNAME)
 
@@ -104,6 +107,7 @@ PERIODIC_INTERVAL = 50
 time_base = 0.0
 time_mult = 1.0
 
+
 def init_time():
     warp = os.environ.get("RENPY_TIMEWARP", "1.0")
 
@@ -113,6 +117,7 @@ def init_time():
     time_base = time.time()
     time_mult = float(warp)
 
+
 def get_time():
     t = time.time()
     return time_base + (t - time_base) * time_mult
@@ -125,6 +130,7 @@ def displayable_by_tag(layer, tag):
 
     return renpy.game.context().scene_lists.get_displayable_by_tag(layer, tag)
 
+
 class IgnoreEvent(Exception):
     """
     Exception that is raised when we want to ignore an event, but
@@ -133,6 +139,7 @@ class IgnoreEvent(Exception):
 
     pass
 
+
 class EndInteraction(Exception):
     """
     Exception that can be raised (for example, during the render method of
@@ -142,6 +149,7 @@ class EndInteraction(Exception):
     def __init__(self, value):
         self.value = value
 
+
 class absolute(float):
     """
     This represents an absolute float coordinate.
@@ -198,6 +206,39 @@ def place(width, height, sw, sh, placement):
 
     return x, y
 
+
+class DisplayableArguments(renpy.object.Object):
+    """
+    Represents a set of arguments that can be passed to a duplicated
+    displayable.
+    """
+
+    # The name of the displayable without any arguments.
+    name = ()
+
+    # Arguments supplied.
+    args = ()
+
+    # This gets set to true if the arguments are consumed.
+    consumed = False
+
+    # The style prefix in play. This is used by DynamicImage to figure
+    # out the prefix list to apply.
+    prefix = None
+
+    def copy(self, **kwargs):
+        """
+        Returns a copy of this object with the various fields set to the
+        values they were given in kwargs.
+        """
+
+        rv = DisplayableArguments()
+        rv.__dict__.update(self.__dict__)
+        rv.__dict__.update(kwargs)
+
+        return rv
+
+
 class Displayable(renpy.object.Object):
     """
     The base class for every object in Ren'Py that can be
@@ -247,11 +288,82 @@ class Displayable(renpy.object.Object):
     # Does this displayable use the scope?
     _uses_scope = False
 
-    def __init__(self, focus=None, default=False, style='default', **properties):
-        self.style = renpy.style.Style(style, properties) # @UndefinedVariable
+    # Arguments supplied to this displayable.
+    _args = DisplayableArguments()
+
+    # Set to true of the displayable is duplicatable (has a non-trivial
+    # duplicate method), or one of its children is.
+    _duplicatable = False
+
+    # Does this displayable require clipping?
+    _clipping = False
+
+    def __init__(self, focus=None, default=False, style='default', _args=None, **properties):
+        self.style = renpy.style.Style(style, properties)  # @UndefinedVariable
         self.focus_name = focus
         self.default = default
 
+        if _args is not None:
+            self._args = _args
+
+    def _copy(self, args=None):
+        """
+        Makes a shallow copy of the displayable. If `args` is provided,
+        replaces the arguments with the stored copy.
+        """
+
+        rv = copy.copy(self)
+
+        if args is not None:
+            rv._args = args
+
+        return rv
+
+    def _duplicate(self, args):
+        """
+        Makes a duplicate copy of the following kids of displayables:
+
+        * Displayables that can accept arguments.
+        * Displayables that maintain state that should be reset before being
+          shown to the user.
+        * Containers that contain (including transitively) one of the other
+          kinds of displayables.
+
+        Displayables that contain state that can be manipulated by the user
+        are never copied.
+
+        This should call _unique on children that have been copied before
+        setting its own _duplicatable flag.
+        """
+
+        return self
+
+    def _in_current_store(self):
+        """
+        Returns a version of this displayable that will not change as it is
+        rendered.
+        """
+
+        return self
+
+    def _unique(self):
+        """
+        This is called when a displayable is "born" unique, which occurs
+        when there is only a single reference to it. What it does is to
+        manage the _duplicatable flag - setting it false unless one of
+        the displayable's children happens to be duplicatable.
+        """
+
+        return
+
+    def parameterize(self, name, parameters):
+        """
+        Obsolete alias for _duplicate.
+        """
+
+        a = self._args.copy(name=name, args=parameters)
+        return self._duplicate(a)
+
     def _equals(self, o):
         """
         This is a utility method that can be called by a Displayable's
@@ -294,7 +406,6 @@ class Displayable(renpy.object.Object):
 
             i.find_focusable(callback, focus_name)
 
-
     def focus(self, default=False):
         """
         Called to indicate that this widget has the focus.
@@ -303,8 +414,7 @@ class Displayable(renpy.object.Object):
         self.set_style_prefix(self.role + "hover_", True)
 
         if not default:
-            if self.style.hover_sound:
-                renpy.audio.music.play(self.style.hover_sound, channel="sound")
+            renpy.exports.play(self.style.hover_sound)
 
     def unfocus(self, default=False):
         """
@@ -335,18 +445,6 @@ class Displayable(renpy.object.Object):
         self.style.set_prefix(prefix)
         renpy.display.render.redraw(self, 0)
 
-    def parameterize(self, name, parameters):
-        """
-        Called to parameterize this. By default, we don't take any
-        parameters.
-        """
-
-        if parameters:
-            raise Exception("Image '%s' can't take parameters '%s'. (Perhaps you got the name wrong?)" %
-                            (' '.join(name), ' '.join(parameters)))
-
-        return self
-
     def render(self, width, height, st, at):
         """
         Called to display this displayable. This is called with width
@@ -495,11 +593,11 @@ class Displayable(renpy.object.Object):
         Called when the displayable is added to a scene list.
         """
 
-    def _get_parameterized(self):
+    def _target(self):
         """
-        If this is a ImageReference to a parameterized image, return
-        the get_parameterized() of the parameterized image. Otherwise,
-        return this displayable.
+        If this displayable is part of a chain of one or more references,
+        returns the ultimate target of those references. Otherwise, returns
+        the displayable.
         """
 
         return self
@@ -520,23 +618,20 @@ class Displayable(renpy.object.Object):
 
         return
 
-    def _in_old_scene(self):
-        """
-        Returns a version of this displayable that will not change as it is
-        rendered.
-        """
-
-        return self
-
     def _tts_common(self, default_alt=None):
 
         rv = [ ]
 
         for i in self.visit():
             if i is not None:
-                rv.append(i._tts())
+                speech = i._tts()
 
-        rv = " ".join(rv)
+                if speech.strip():
+                    rv.append(speech)
+
+        rv = ": ".join(rv)
+        rv = rv.replace("::", ":")
+        rv = rv.replace(": :", ":")
 
         alt = self.style.alt
 
@@ -552,12 +647,11 @@ class Displayable(renpy.object.Object):
         """
         Returns the self-voicing text of this displayable and all of its
         children that cannot take focus. If the displayable can take focus,
-        retuns the empty string.
+        returns the empty string.
         """
 
         return self._tts_common()
 
-
     def _tts_all(self):
         """
         Returns the self-voicing text of this displayable and all of its
@@ -757,18 +851,17 @@ class SceneLists(renpy.object.Object):
         if isinstance(new_thing, renpy.display.screen.ScreenDisplayable):
             return new_thing
 
-        old_transform = old_thing._get_parameterized()
+        old_transform = old_thing._target()
         if not isinstance(old_transform, renpy.display.motion.Transform):
             return new_thing
 
-        new_transform = new_thing._get_parameterized()
+        new_transform = new_thing._target()
         if not isinstance(new_transform, renpy.display.motion.Transform):
             new_thing = new_transform = renpy.display.motion.Transform(child=new_thing)
 
         new_transform.take_state(old_transform)
         return new_thing
 
-
     def find_index(self, layer, tag, zorder, behind):
         """
         This finds the spot in the named layer where we should insert the
@@ -791,7 +884,7 @@ class SceneLists(renpy.object.Object):
                         zorder = sle.zorder
 
         if zorder is None:
-            zorder = 0
+            zorder = renpy.config.tag_zorder.get(tag, 0)
 
         for i, sle in enumerate(self.layers[layer]):
 
@@ -809,7 +902,6 @@ class SceneLists(renpy.object.Object):
 
         return add_index, remove_index, zorder
 
-
     def add(self,
             layer,
             thing,
@@ -888,9 +980,9 @@ class SceneLists(renpy.object.Object):
                 st = sle.show_time
 
             if (not atl and
-                not at_list and
-                renpy.config.keep_running_transform and
-                isinstance(old, renpy.display.motion.Transform)):
+                    not at_list and
+                    renpy.config.keep_running_transform and
+                    isinstance(old, renpy.display.motion.Transform)):
 
                 thing = sle.displayable._change_transform_child(thing)
             else:
@@ -1003,7 +1095,7 @@ class SceneLists(renpy.object.Object):
         if layer not in self.layers:
             raise Exception("Trying to remove something from non-existent layer '%s'." % layer)
 
-        _add_index, remove_index, zorder = self.find_index(layer, thing, 0, [ ])
+        _add_index, remove_index, _zorder = self.find_index(layer, thing, 0, [ ])
 
         if remove_index is not None:
             tag = self.layers[layer][remove_index].tag
@@ -1022,7 +1114,7 @@ class SceneLists(renpy.object.Object):
         totally wiped out.
         """
 
-        if not layer in self.layers:
+        if layer not in self.layers:
             return
 
         if not hide:
@@ -1048,7 +1140,7 @@ class SceneLists(renpy.object.Object):
         time with the given time.
         """
 
-        for l, (t, list) in self.layer_at_list.items(): #@ReservedAssignment
+        for l, (t, list) in self.layer_at_list.items():  # @ReservedAssignment
             self.layer_at_list[l] = (t or time, list)
 
         for l, ll in self.layers.iteritems():
@@ -1073,6 +1165,7 @@ class SceneLists(renpy.object.Object):
         rv = renpy.display.layout.MultiBox(layout='fixed', focus=layer, **properties)
         rv.append_scene_list(self.layers[layer])
         rv.layer_name = layer
+        rv._duplicatable = False
 
         time, at_list = self.layer_at_list[layer]
 
@@ -1091,7 +1184,6 @@ class SceneLists(renpy.object.Object):
 
             rv = f
 
-
         return rv
 
     def remove_hide_replaced(self, layer, tag):
@@ -1136,6 +1228,25 @@ class SceneLists(renpy.object.Object):
 
             self.layers[l] = newl
 
+    def remove_all_hidden(self):
+        """
+        Removes everything hidden, even if it's not time yet. (Used when making a rollback copy).
+        """
+
+        for l in self.layers:
+            newl = [ ]
+
+            for sle in self.layers[l]:
+
+                if sle.tag:
+
+                    if "$" in sle.tag:
+                        continue
+
+                newl.append(sle)
+
+            self.layers[l] = newl
+
     def get_displayable_by_tag(self, layer, tag):
         """
         Returns the displayable on the layer with the given tag, or None
@@ -1253,6 +1364,7 @@ class MouseMove(object):
         renpy.display.draw.set_mouse_pos(x, y)
         return True
 
+
 def get_safe_mode():
     """
     Returns true if we should go into safe mode.
@@ -1279,6 +1391,11 @@ def get_safe_mode():
     except:
         return False
 
+
+# How long should we be in maximum framerate mode at the start of the game?
+initial_maximum_framerate = 0.0
+
+
 class Interface(object):
     """
     This represents the user interface that interacts with the user.
@@ -1392,21 +1509,20 @@ class Interface(object):
             if layer in renpy.config.layer_clipping:
                 x, y, w, h = renpy.config.layer_clipping[layer]
                 self.layer_properties[layer] = dict(
-                    xpos = x,
-                    xanchor = 0,
-                    ypos = y,
-                    yanchor = 0,
-                    xmaximum = w,
-                    ymaximum = h,
-                    xminimum = w,
-                    yminimum = h,
-                    clipping = True,
+                    xpos=x,
+                    xanchor=0,
+                    ypos=y,
+                    yanchor=0,
+                    xmaximum=w,
+                    ymaximum=h,
+                    xminimum=w,
+                    yminimum=h,
+                    clipping=True,
                     )
 
             else:
                 self.layer_properties[layer] = dict()
 
-
         # A stack giving the values of self.transition and self.transition_time
         # for contexts outside the current one. This is used to restore those
         # in the case where nothing has changed in the new context.
@@ -1430,13 +1546,14 @@ class Interface(object):
         # The thread that can do display operations.
         self.thread = threading.current_thread()
 
-        # Initialize pygame.
-        if pygame.version.vernum < (1, 8, 1):
-            raise Exception("Ren'Py requires pygame 1.8.1 to run.")
-
-        # pygame.font.init()
+        # Initialize audio.
         renpy.audio.audio.init()
-        pygame.display.init()
+
+        # Initialize pygame.
+        try:
+            pygame.display.init()
+        except:
+            pass
 
         # Init timing.
         init_time()
@@ -1495,6 +1612,60 @@ class Interface(object):
         self.safe_mode = get_safe_mode()
         renpy.safe_mode_checked = True
 
+        # A scale factor used to compensate for the system DPI.
+        self.dpi_scale = self.setup_dpi_scaling()
+
+        renpy.display.log.write("DPI scale factor: %f", self.dpi_scale)
+
+        # A time until which we should draw at maximum framerate.
+        self.maximum_framerate_time = 0.0
+        self.maximum_framerate(initial_maximum_framerate)
+
+        # True if this is the first interact.
+        self.start_interact = True
+
+    def setup_dpi_scaling(self):
+
+        if "RENPY_HIGHDPI" in os.environ:
+            return float(os.environ["RENPY_HIGHDPI"])
+
+        if not renpy.windows:
+            return 1.0
+
+        try:
+            import ctypes
+            from ctypes import c_void_p, c_int
+
+            ctypes.windll.user32.SetProcessDPIAware()
+
+            GetDC = ctypes.windll.user32.GetDC
+            GetDC.restype = c_void_p
+            GetDC.argtypes = [ c_void_p ]
+
+            ReleaseDC = ctypes.windll.user32.ReleaseDC
+            ReleaseDC.argtypes = [ c_void_p, c_void_p ]
+
+            GetDeviceCaps = ctypes.windll.gdi32.GetDeviceCaps
+            GetDeviceCaps.restype = c_int
+            GetDeviceCaps.argtypes = [ c_void_p, c_int ]
+
+            LOGPIXELSX = 88
+
+            dc = GetDC(None)
+            rv = GetDeviceCaps(dc, LOGPIXELSX) / 96.0
+            ReleaseDC(None, dc)
+
+            if rv < renpy.config.de_minimus_dpi_scale:
+                renpy.display.log.write("De minimus DPI scale, was %r", rv)
+                rv = 1.0
+
+            return rv
+
+        except:
+            renpy.display.log.write("Could not determine DPI scale factor:")
+            renpy.display.log.exception()
+            return 1.0
+
     def start(self):
         """
         Starts the interface, by opening a window and setting the mode.
@@ -1503,6 +1674,8 @@ class Interface(object):
         if self.started:
             return
 
+        renpy.display.render.render_ready()
+
         # Kill off the presplash.
         renpy.display.presplash.end()
 
@@ -1529,7 +1702,7 @@ class Interface(object):
         renpy.display.log.write(s)
 
         if renpy.android and not renpy.config.log_to_stdout:
-            print s
+            print(s)
 
     def post_init(self):
         """
@@ -1538,7 +1711,7 @@ class Interface(object):
 
         # Needed for Unity.
         wmclass = renpy.config.save_directory or os.path.basename(sys.argv[0])
-        os.environ['SDL_VIDEO_X11_WMCLASS'] = wmclass
+        os.environ[b'SDL_VIDEO_X11_WMCLASS'] = wmclass.encode("utf-8")
 
         self.set_window_caption(force=True)
         self.set_icon()
@@ -1561,8 +1734,6 @@ class Interface(object):
 
             pygame.event.set_blocked(i)
 
-
-
     def set_icon(self):
         """
         This is called to set up the window icon.
@@ -1571,15 +1742,11 @@ class Interface(object):
         # Window icon.
         icon = renpy.config.window_icon
 
-        if renpy.windows and renpy.config.windows_icon:
-            icon = renpy.config.windows_icon
-
         if icon:
 
             im = renpy.display.scale.image_load_unscaled(
                 renpy.loader.load(icon),
                 icon,
-                convert=False,
                 )
 
             # Convert the aspect ratio to be square.
@@ -1589,14 +1756,16 @@ class Interface(object):
             square_im.blit(im, ( (imax-iw)/2, (imax-ih)/2 ))
             im = square_im
 
-            if renpy.windows and im.get_size() != (32, 32):
-                im = renpy.display.scale.real_smoothscale(im, (32, 32))
-
             pygame.display.set_icon(im)
 
-
     def set_window_caption(self, force=False):
-        caption = renpy.config.window_title + renpy.store._window_subtitle
+
+        window_title = renpy.config.window_title
+
+        if window_title is None:
+            window_title = "A Ren'Py Game"
+
+        caption = renpy.translation.translate_string(window_title) + renpy.store._window_subtitle
 
         if renpy.exports.get_autoreload():
             caption += " - autoreload"
@@ -1671,7 +1840,6 @@ class Interface(object):
 
         return rv
 
-
     def kill_textures(self):
         renpy.display.render.free_memory()
         renpy.text.text.layout_cache_clear()
@@ -1697,8 +1865,8 @@ class Interface(object):
 
         if self.display_reset:
 
-            pygame.key.stop_text_input() # @UndefinedVariable
-            pygame.key.set_text_input_rect(None) # @UndefinedVariable
+            pygame.key.stop_text_input()  # @UndefinedVariable
+            pygame.key.set_text_input_rect(None)  # @UndefinedVariable
             self.text_rect = None
 
             renpy.display.draw.deinit()
@@ -1724,7 +1892,7 @@ class Interface(object):
         virtual_size = (renpy.config.screen_width, renpy.config.screen_height)
 
         if physical_size is None:
-            if renpy.mobile or renpy.game.preferences.physical_size is None: #@UndefinedVariable
+            if renpy.mobile or renpy.game.preferences.physical_size is None:  # @UndefinedVariable
                 physical_size = (None, None)
             else:
                 physical_size = renpy.game.preferences.physical_size
@@ -1779,7 +1947,6 @@ class Interface(object):
         # True if we're doing a one-time profile.
         self.profile_once = False
 
-
     def draw_screen(self, root_widget, fullscreen_video, draw):
 
         surftree = renpy.display.render.render_screen(
@@ -1797,7 +1964,6 @@ class Interface(object):
         self.surftree = surftree
         self.fullscreen_video = fullscreen_video
 
-
     def take_screenshot(self, scale, background=False):
         """
         This takes a screenshot of the current screen, and stores it so
@@ -1837,7 +2003,6 @@ class Interface(object):
         self.screenshot = sio.getvalue()
         sio.close()
 
-
     def check_background_screenshot(self):
         """
         Handles requests for a background screenshot.
@@ -1869,7 +2034,6 @@ class Interface(object):
 
         return rv
 
-
     def lose_screenshot(self):
         """
         This deallocates the saved screenshot.
@@ -1878,7 +2042,6 @@ class Interface(object):
         self.screenshot = None
         self.screenshot_surface = None
 
-
     def save_screenshot(self, filename):
         """
         Saves a full-size screenshot in the given filename.
@@ -1896,8 +2059,6 @@ class Interface(object):
                 raise
             pass
 
-
-
     def show_window(self):
 
         if not renpy.store._window:
@@ -1910,7 +2071,14 @@ class Interface(object):
             return
 
         if renpy.config.empty_window:
-            renpy.config.empty_window()
+
+            old_history = renpy.store._history  # @UndefinedVariable
+            renpy.store._history = False
+
+            try:
+                renpy.config.empty_window()
+            finally:
+                renpy.store._history = old_history
 
     def do_with(self, trans, paired, clear=False):
 
@@ -1956,7 +2124,6 @@ class Interface(object):
         scene_lists.replace_transient()
         scene_lists.shown_window = False
 
-
     def set_transition(self, transition, layer=None, force=False):
         """
         Sets the transition that will be performed as part of the next
@@ -1971,7 +2138,6 @@ class Interface(object):
         else:
             self.transition[layer] = transition
 
-
     def event_peek(self):
         """
         This peeks the next event. It returns None if no event exists.
@@ -2007,7 +2173,6 @@ class Interface(object):
 
         return rv
 
-
     def event_wait(self):
         """
         This is in its own function so that we can track in the
@@ -2044,7 +2209,6 @@ class Interface(object):
 
         renpy.ui.close()
 
-
     def compute_scene(self, scene_lists):
         """
         This converts scene lists into a dictionary mapping layer
@@ -2066,7 +2230,6 @@ class Interface(object):
 
         return rv
 
-
     def quit_event(self):
         """
         This is called to handle the user invoking a quit.
@@ -2094,10 +2257,9 @@ class Interface(object):
         else:
             raise renpy.game.QuitException()
 
-
     def get_mouse_info(self):
         # Figure out if the mouse visibility algorithm is hiding the mouse.
-        if self.mouse_event_time + renpy.config.mouse_hide_time < renpy.display.core.get_time():
+        if (renpy.config.mouse_hide_time is not None) and (self.mouse_event_time + renpy.config.mouse_hide_time < renpy.display.core.get_time()):
             visible = False
         else:
             visible = renpy.store.mouse_visible and (not renpy.game.less_mouse)
@@ -2182,7 +2344,7 @@ class Interface(object):
 
         renpy.exports.free_memory()
 
-        print "Entered background."
+        print("Entered background.")
 
         while True:
             ev = pygame.event.wait()
@@ -2193,7 +2355,7 @@ class Interface(object):
             if ev.type == pygame.APP_TERMINATING:
                 sys.exit(0)
 
-        print "Entering foreground."
+        print("Entering foreground.")
 
         # Since we came back to life, we can get rid of the
         # auto-reload.
@@ -2224,7 +2386,6 @@ class Interface(object):
 
         renpy.display.log.write("The window was minimized.")
 
-
     def restored(self):
         """
         Called when we are restored from being an icon.
@@ -2272,7 +2433,6 @@ class Interface(object):
         self.ignore_touch = True
         renpy.display.focus.mouse_handler(None, -1, -1, default=False)
 
-
     def text_event_in_queue(self):
         """
         Returns true if the next event in the queue is a text editing event.
@@ -2289,15 +2449,15 @@ class Interface(object):
         Updates the text input state and text rectangle.
         """
 
-        if renpy.store._text_rect is not None: # @UndefinedVariable
-            self.text_rect = renpy.store._text_rect # @UndefinedVariable
+        if renpy.store._text_rect is not None:  # @UndefinedVariable
+            self.text_rect = renpy.store._text_rect  # @UndefinedVariable
 
         if self.text_rect is not None:
 
-            not_shown = pygame.key.has_screen_keyboard_support() and not pygame.key.is_screen_keyboard_shown() # @UndefinedVariable
+            not_shown = pygame.key.has_screen_keyboard_support() and not pygame.key.is_screen_keyboard_shown()  # @UndefinedVariable
 
             if not self.old_text_rect or not_shown:
-                pygame.key.start_text_input() # @UndefinedVariable
+                pygame.key.start_text_input()  # @UndefinedVariable
 
             if self.old_text_rect != self.text_rect:
                 x, y, w, h = self.text_rect
@@ -2305,17 +2465,26 @@ class Interface(object):
                 x1, y1 = renpy.display.draw.untranslate_point(x + w, y + h)
                 rect = (x0, y0, x1 - x0, y1 - y0)
 
-                pygame.key.set_text_input_rect(rect) # @UndefinedVariable
+                pygame.key.set_text_input_rect(rect)  # @UndefinedVariable
 
         else:
             if self.old_text_rect:
-                pygame.key.stop_text_input() # @UndefinedVariable
-                pygame.key.set_text_input_rect(None) # @UndefinedVariable
+                pygame.key.stop_text_input()  # @UndefinedVariable
+                pygame.key.set_text_input_rect(None)  # @UndefinedVariable
 
         self.old_text_rect = self.text_rect
 
+    def maximum_framerate(self, t):
+        """
+        Forces Ren'Py to draw the screen at the maximum framerate for `t` seconds.
+        """
 
-    def interact(self, clear=True, suppress_window=False, **kwargs):
+        if t is None:
+            self.maximum_framerate_time = 0
+        else:
+            self.maximum_framerate_time = max(self.maximum_framerate_time, get_time() + t)
+
+    def interact(self, clear=True, suppress_window=False, trans_pause=False, **kwargs):
         """
         This handles an interaction, restarting it if necessary. All of the
         keyword arguments are passed off to interact_core.
@@ -2324,6 +2493,8 @@ class Interface(object):
         if not self.started:
             self.start()
 
+        self.trans_pause = trans_pause
+
         # Cancel magic error reporting.
         renpy.bootstrap.report_error = None
 
@@ -2334,7 +2505,6 @@ class Interface(object):
 
         context.interacting = True
 
-
         # Show a missing window.
         if not suppress_window:
             self.show_window()
@@ -2350,8 +2520,11 @@ class Interface(object):
 
             repeat = True
 
+            self.start_interact = True
+
             while repeat:
-                repeat, rv = self.interact_core(preloads=preloads, **kwargs)
+                repeat, rv = self.interact_core(preloads=preloads, trans_pause=trans_pause, **kwargs)
+                self.start_interact = False
 
             return rv
 
@@ -2381,7 +2554,6 @@ class Interface(object):
                       preloads=[],
                       roll_forward=None,
                       ):
-
         """
         This handles one cycle of displaying an image to the user,
         and then responding to user input.
@@ -2396,6 +2568,16 @@ class Interface(object):
         @param suppress_underlay: This suppresses the display of the underlay.
         """
 
+        suppress_overlay = suppress_overlay or renpy.store.suppress_overlay
+
+        # Store the various parameters.
+        self.suppress_overlay = suppress_overlay
+        self.suppress_underlay = suppress_underlay
+        self.trans_pause = trans_pause
+
+        # Show default screens.
+        renpy.display.screen.show_overlay_screens(suppress_overlay)
+
         # Prepare screens, if need be.
         renpy.display.screen.prepare_screens()
 
@@ -2413,7 +2595,7 @@ class Interface(object):
                 continue
 
             self.ongoing_transition[k] = self.transition[k]
-            self.transition_from[k] = self.old_scene[k]._in_old_scene()
+            self.transition_from[k] = self.old_scene[k]._in_current_store()
             self.transition_time[k] = None
 
         self.transition.clear()
@@ -2423,7 +2605,7 @@ class Interface(object):
             self.transition_from.clear()
             self.transition_time.clear()
 
-        ## Safety condition, prevents deadlocks.
+        # Safety condition, prevents deadlocks.
         if trans_pause:
             if not self.ongoing_transition:
                 return False, None
@@ -2525,7 +2707,7 @@ class Interface(object):
             focus_roots.append(scene_layer)
 
             if (self.ongoing_transition.get(layer, None) and
-                not suppress_transition):
+                    not suppress_transition):
 
                 trans = self.ongoing_transition[layer](
                     old_widget=self.transition_from[layer],
@@ -2549,7 +2731,7 @@ class Interface(object):
 
         # Add layers_root to root_widget, perhaps through a transition.
         if (self.ongoing_transition.get(None, None) and
-            not suppress_transition):
+                not suppress_transition):
 
             old_root = renpy.display.layout.MultiBox(layout='fixed')
             old_root.layers = { }
@@ -2607,7 +2789,9 @@ class Interface(object):
         renpy.display.video.early_interact()
 
         # Call per-interaction code for all widgets.
+        renpy.display.behavior.input_pre_per_interact()
         root_widget.visit_all(lambda i : i.per_interact())
+        renpy.display.behavior.input_post_per_interact()
 
         # Now, update various things regarding scenes and transitions,
         # so we are ready for a new interaction or a restart.
@@ -2658,7 +2842,9 @@ class Interface(object):
                 renpy.execution.not_infinite_loop(10)
 
                 # Check for a change in fullscreen preference.
-                if self.fullscreen != renpy.game.preferences.fullscreen or self.display_reset:
+                if ((self.fullscreen != renpy.game.preferences.fullscreen) or
+                        self.display_reset or (renpy.display.draw is None)):
+
                     self.set_mode()
                     needs_redraw = True
 
@@ -2667,6 +2853,10 @@ class Interface(object):
                     renpy.loader.needs_autoreload = False
                     renpy.exports.reload_script()
 
+                for i in renpy.config.needs_redraw_callbacks:
+                    if i():
+                        needs_redraw = True
+
                 # Redraw the screen.
                 if (self.force_redraw or
                     ((first_pass or not pygame.event.peek(ALL_EVENTS)) and
@@ -2688,6 +2878,8 @@ class Interface(object):
                     if not self.interact_time:
                         self.interact_time = self.frame_time
 
+                    renpy.audio.audio.advance_time()  # Sets the time of all video frames.
+
                     self.draw_screen(root_widget, fullscreen_video, (not fullscreen_video) or video_frame_drawn)
 
                     if first_pass:
@@ -2703,25 +2895,25 @@ class Interface(object):
                         new_time = get_time()
 
                         if self.profile_once or (new_time - self.profile_time > .015):
-                            print "Profile: Redraw took %.3f ms." % (1000 * (new_time - self.frame_time))
-                            print "Profile: %.3f ms to complete event." % (1000 * (new_time - self.profile_time))
+                            print("Profile: Redraw took %.3f ms." % (1000 * (new_time - self.frame_time)))
+                            print("Profile: %.3f ms to complete event." % (1000 * (new_time - self.profile_time)))
 
                         self.profile_once = False
 
                     if first_pass and self.last_event and self.last_event.type in [ pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP, pygame.MOUSEMOTION ]:
 
-                            x, y = renpy.display.draw.get_mouse_pos()
-                            ev, x, y = renpy.display.emulator.emulator(self.last_event, x, y)
+                        x, y = renpy.display.draw.get_mouse_pos()
+                        ev, x, y = renpy.display.emulator.emulator(self.last_event, x, y)
 
-                            if self.ignore_touch:
-                                x = -1
-                                y = -1
+                        if self.ignore_touch:
+                            x = -1
+                            y = -1
 
-                            if renpy.android and self.last_event.type == pygame.MOUSEBUTTONUP:
-                                x = -1
-                                y = -1
+                        if renpy.android and self.last_event.type == pygame.MOUSEBUTTONUP:
+                            x = -1
+                            y = -1
 
-                            renpy.display.focus.mouse_handler(None, x, y, default=False)
+                        renpy.display.focus.mouse_handler(None, x, y, default=False)
 
                     needs_redraw = False
                     first_pass = False
@@ -2732,6 +2924,8 @@ class Interface(object):
 
                     self.update_text_rect()
 
+                    renpy.test.testexecution.execute()
+
                 # Move the mouse, if necessary.
                 if self.mouse_move is not None:
                     if not self.mouse_move.perform():
@@ -2754,6 +2948,9 @@ class Interface(object):
                 if renpy.display.render.process_redraws():
                     needs_redraw = True
 
+                if self.maximum_framerate_time > get_time():
+                    needs_redraw = True
+
                 # How many seconds until we timeout.
                 _timeout_in = 3600
 
@@ -2831,6 +3028,10 @@ class Interface(object):
                     ev = self.event_wait()
 
                 if ev.type == pygame.NOEVENT:
+
+                    if not needs_redraw or self.mouse_move:
+                        pygame.time.wait(1)
+
                     continue
 
                 self.profile_time = get_time()
@@ -2910,24 +3111,25 @@ class Interface(object):
                     if ev.w == 1 and ev.h == 1:
                         continue
 
+                    size = (ev.w // self.dpi_scale, ev.h // self.dpi_scale)
+
                     if pygame.display.get_surface().get_size() != ev.size:
-                        self.set_mode((ev.w, ev.h))
+                        self.set_mode(size)
 
                     if not self.fullscreen:
-                        self.last_resize = ev.size
+                        self.last_resize = size
 
                     continue
 
                 # If we're ignoring touch events, and get a mouse up, stop
                 # ignoring those events.
                 if self.ignore_touch and \
-                    ev.type == pygame.MOUSEBUTTONUP and \
-                    ev.button == 1:
+                        ev.type == pygame.MOUSEBUTTONUP and \
+                        ev.button == 1:
 
                     self.ignore_touch = False
                     continue
 
-
                 # Merge mousemotion events.
                 if ev.type == pygame.MOUSEMOTION:
                     evs = pygame.event.get([pygame.MOUSEMOTION])
@@ -2951,6 +3153,9 @@ class Interface(object):
                 if ev.type == pygame.ACTIVEEVENT:
 
                     if ev.state & 1:
+                        if not ev.gain:
+                            renpy.display.focus.clear_focus()
+
                         self.mouse_focused = ev.gain
 
                     if ev.state & 2:
@@ -2967,6 +3172,7 @@ class Interface(object):
                 # This returns the event location. It also updates the
                 # mouse state as necessary.
                 x, y = renpy.display.draw.mouse_event(ev)
+                x, y = renpy.test.testmouse.get_mouse_pos(x, y)
 
                 ev, x, y = renpy.display.emulator.emulator(ev, x, y)
                 if ev is None:
@@ -2989,7 +3195,7 @@ class Interface(object):
                 try:
 
                     if self.touch:
-                        renpy.display.gesture.recognizer.event(ev, x, y) # @UndefinedVariable
+                        renpy.display.gesture.recognizer.event(ev, x, y)  # @UndefinedVariable
 
                     # Handle the event normally.
                     rv = renpy.display.focus.mouse_handler(ev, x, y)
@@ -3012,7 +3218,6 @@ class Interface(object):
                             l = self.surftree.main_displayables_at_point(x, y, renpy.config.layers)
                             renpy.game.invoke_in_new_context(renpy.config.inspector, l)
 
-
                 except IgnoreEvent:
                     # An ignored event can change the timeout. So we want to
                     # process an TIMEEVENT to ensure that the timeout is
@@ -3021,7 +3226,6 @@ class Interface(object):
                     if ev.type != TIMEEVENT:
                         self.post_time_event()
 
-
                 # Check again after handling the event.
                 needs_redraw |= renpy.display.render.process_redraws()
 
